#!/bin/sh
# Copyright 2013 IBM, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

VDSM_TOOL="/usr/bin/vdsm-tool"
UPGRADE_LOGGING_PARAMS="--vvverbose --append --logfile=/var/log/vdsm/upgrade.log"
prog=vdsm

#### pre-start tasks ####
_dev_has_ipv4() {
    local dev="$1"
    ip -o address show "$dev" | grep -q ' inet '
}

_all_devs_have_ipv4(){
    local dev

    for dev in $1
    do
        _dev_has_ipv4 "$dev" || return 1
    done
    return 0
}

_wait_for_ipv4() {
    # wait up to a timeout for all specified devices to have an ipv4 address
    local i

    for i in `seq 10`
    do
        _all_devs_have_ipv4 "$1" && return 0
        sleep 1
    done
    echo "timeout waiting for ipv4 addresses"
    return 1
}

task_wait_for_network() {
    # interfaces that we depend upon may not have their IP address ready yet
    # see https://bugzilla.redhat.com/show_bug.cgi?id=1174611#c34
    local interfaces dev static_devs

    pushd /etc/sysconfig/network-scripts/ > /dev/null

    # copied from /etc/init.d/network
    interfaces=$(ls ifcfg-* | \
            LC_ALL=C sed -e "$__sed_discard_ignored_files" \
                   -e '/\(ifcfg-lo$\|:\|ifcfg-.*-range\)/d' \
                   -e '{ s/^ifcfg-//g;s/[0-9]/ &/}' | \
            LC_ALL=C sort -k 1,1 -k 2n | \
            LC_ALL=C sed 's/ //')

    for dev in $interfaces
    do
        ifcfg="ifcfg-$dev"
        grep -q '# Generated by VDSM version' "$ifcfg" || continue
        grep -q 'ONBOOT=yes' "$ifcfg" || continue
        grep -q 'IPADDR=' "$ifcfg" || continue
        static_devs="$static_devs $dev"
    done
    popd > /dev/null

    _wait_for_ipv4 "$static_devs"

    # never fail, in order to allow intentionally-downed vdsm-controlled
    # interfaces
    return 0
}

task_configure_coredump() {
    local conf_file="/etc/vdsm/vdsm.conf"
    local getconfitem="/usr/share/vdsm/get-conf-item"

    if "${getconfitem}" "${conf_file}" vars core_dump_enable false |
      tr A-Z a-z | grep -q true; then
        echo "/var/log/core/core.%p.%t.dump" > /proc/sys/kernel/core_pattern
    fi
}

task_configure_vdsm_logs() {
    local vdsm_logs="
        /var/log/vdsm/connectivity.log
        /var/log/vdsm/mom.log
        /var/log/vdsm/vdsm.log
    "

    for logfile in ${vdsm_logs}; do
        if [ ! -e "${logfile}" ]; then
            /bin/touch "${logfile}"
            /bin/chmod 0644 "${logfile}"
            /bin/chown vdsm:kvm "${logfile}"
        fi
    done
}

task_run_init_hooks(){
    "/usr/bin/python" "/usr/share/vdsm/hooks.pyc" before_vdsm_start
}


task_check_is_configured() {
    "$VDSM_TOOL" is-configured
}


task_prepare_transient_repository() {
    "$VDSM_TOOL" setup-transient-repository
    "$VDSM_TOOL" cleanup-transient-repository
}


task_syslog_available() {
    if ! [ -S "/dev/log" ]; then
        echo "    Cannot communicate with syslog daemon for reports. "
        echo "    Install syslog daemon provider, configure it properly"
        echo "    and start the service."
        echo "    (Alternatives: rsyslog, syslog-ng)"
        return 1
    fi
    return 0
}


task_nwfilter(){
    "${VDSM_TOOL}" nwfilter
}


task_dummybr(){
    "${VDSM_TOOL}" dummybr
}


_has_systemd() {
    "/bin/mountpoint" -q /cgroup/systemd ||
        "/bin/mountpoint" -q /sys/fs/cgroup/systemd
}


task_load_needed_modules(){
    if ! _has_systemd; then
        "${VDSM_TOOL}" load-needed-modules
    fi
}


task_dump_bonding_defaults(){
    "${VDSM_TOOL}" dump-bonding-defaults
}


task_tune_system(){
    "/sbin/sysctl" -q -p "/etc/sysctl.d/vdsm.conf"
}


_CONF_FILE="/etc/vdsm/vdsm.conf"
_GETCONFITEM="/usr/share/vdsm/get-conf-item"
_DOM_METADATA_BACKUP_DIR="/var/log/vdsm/backup"

_mk_data_center() {
    local dc
    dc="$("${_GETCONFITEM}" "${_CONF_FILE}" irs repository /rhev/)"
    /bin/mkdir -p "${dc}"
    "/bin/chown" vdsm:kvm "${dc}"
}

_mk_dom_backup() {
    /bin/mkdir -p "${_DOM_METADATA_BACKUP_DIR}" > /dev/null 2>&1
    "/bin/chown" vdsm:kvm "$(dirname "${_DOM_METADATA_BACKUP_DIR}")" > /dev/null 2>&1
    "/bin/chown" vdsm:kvm "${_DOM_METADATA_BACKUP_DIR}" > /dev/null 2>&1
}

_mk_core_path() {
    local core_path=/var/log/core
    if ! [ -d "${core_path}" ]; then
        /bin/mkdir -p "${core_path}" > /dev/null 2>&1
    fi
    "/bin/chmod" a+tw "${core_path}" > /dev/null 2>&1
}

_mk_run_path() {
    local run_path

    for run_path in "/var/run/vdsm"; do
        /bin/mkdir -p "${run_path}"
        "/bin/chmod" 755 "${run_path}"
        "/bin/chown" vdsm:kvm "${run_path}"
        "/sbin/restorecon" "${run_path}"
    done
}

task_mkdirs(){
    _mk_data_center
    _mk_core_path
    _mk_dom_backup
    _mk_run_path
    "/bin/chmod" 1777 /dev/shm
}


_free_space() {
    local pathToTest="$1"
    df -P "${pathToTest}" | awk '{print $4}'| tail -1
}

task_test_space() {
    local MIN_SPACE_KB=10000

    if [ "$(_free_space "/var/log/vdsm")" -lt "${MIN_SPACE_KB}" ]; then
        echo "${prog}: low log space"
        return 1
    fi
    return 0
}


task_test_lo() {
    if ! LC_ALL=C "/sbin/ip" link show lo | "/bin/grep" -q UP; then
        echo "VDSMD: lo interface is down, can't run !"
        echo "VDSMD: lo interface is down, can't run !" > /dev/kmsg
        return 1
    fi
    return 0
}


task_validate_configuration(){
    "$VDSM_TOOL" validate-config
}


task_upgraded_version_check(){
    local upgraded_ver_file="/var/lib/vdsm/upgraded_version"
    if [ -f "${upgraded_ver_file}" ]; then
        local ret=0
        if grep -q '^vdsm-4\.10\.' "${upgraded_ver_file}"; then
             # We need to treat the upgrade from 4.10.x specifically
             # by running manual configure command
             "$VDSM_TOOL" configure --force
             ret=$?
        fi
        [ "${ret}" -eq 0 ] && rm -f "${upgraded_ver_file}"
    fi
}
#### pre-start tasks end ####


#### post-stop tasks ####
task_run_final_hooks(){
    "/usr/bin/python" "/usr/share/vdsm/hooks.pyc" after_vdsm_stop
}
#### post-stop tasks end ####


usage(){
    echo "Usage:"
    echo "vdsm_init_common.sh --pre-start|--post-stop"
}

run_tasks(){
    local tasks="$1"
    for taskName in $tasks ; do
        echo "${prog}: Running ${taskName}"
        "task_${taskName}"
        rt=$?
        if [ "$rt" -ne 0 ]; then
            echo "${prog}: stopped during execute ${taskName} task" \
                 "(task returned with error code ${rt})."
            exit 1
        fi
    done
}

case "$1" in
    --pre-start)
        # If dump_bonding_defaults is desired (for the uneliminable possibility
        # of added bonding options or tweaked kernel defaults) it has to be run
        # after load_needed_modules (modprobe bonding).
        # NetworkManager must not crash on nmcli conn delete.
        run_tasks " \
            mkdirs \
            configure_coredump \
            configure_vdsm_logs \
            wait_for_network \
            run_init_hooks \
            upgraded_version_check \
            check_is_configured \
            validate_configuration \
            prepare_transient_repository \
            syslog_available \
            nwfilter \
            dummybr \
            load_needed_modules \
            tune_system \
            test_space \
            test_lo \
            "
        ;;
    --post-stop)
        run_tasks " \
            run_final_hooks \
            "
        ;;
    --help|*)
        usage
        exit 1
        ;;
esac
