#
# Copyright 2011 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import errno
import os
import socket
import traceback

BLANK_UUID = '00000000-0000-0000-0000-000000000000'
DC = '/rhev/data-center'
MASTERSD_PATH = 'mastersd/master/vms'
CDROM_IMAGE = '11111111-1111-1111-1111-111111111111'


class StorageTable:

    def __init__(self, server):
        self.serverConnection = server
        self.columns = ['Vms', 'Domains', 'Images', 'Volumes', 'Template']
        self._buildStorageTable()

    def _getVms(self, pool, vmUUIDs):
        disks = []
        images = []

        for vmUUID in vmUUIDs:
            vm_images_links = os.listdir(os.path.join(DC, pool, MASTERSD_PATH,
                                                      vmUUID))
            for vm_image_link in vm_images_links:
                ovf = open(os.path.join(DC, pool, MASTERSD_PATH, vmUUID,
                                        vm_image_link)).read()
                cont = ovf.split('ovf:fileRef=\"')
                vmName = cont[-1].split('<Name>')[-1].split('</Name>')[0]

                # Avoiding diskless VMs
                if len(cont) < 2:
                    continue

                self.vmList[vmUUID] = {'Name': vmName, 'Refs': []}
                for x in range(1, len(cont)):
                    image = cont[x].split(' ')[0].split('/')[0]
                    if image not in images:
                        s = self.serverConnection
                        sd = s.getImageDomainsList(pool, image)['domainslist']
                        if len(sd) > 0:
                            sd = sd[0]
                            images.append(image)
                            self._setSdInfo(sd)
                            disks.append([sd, image])
                            self.vmList[vmUUID]['Refs'].append([sd, image])

        self._getImages(disks)

    def _setSdInfo(self, sd):

        if sd not in self.domainsList.keys():
            sd_info = self.serverConnection.getStorageDomainInfo(sd)['info']
            self.domainsList[sd] = {'Name': sd_info['name'],
                                    'Role': sd_info['role'], 'Refs': []}

    def _getSds(self, pool):

        disks = []
        sds = self.serverConnection.getStorageDomainsList(pool)
        for sd in sds['domlist']:
            images = []
            imagesList = self.serverConnection.getImagesList(sd)['imageslist']
            self._setSdInfo(sd)

            for image in imagesList:
                images.append(image)
                disks.append([sd, image])
                if image != CDROM_IMAGE:
                    self.domainsList[sd]['Refs'].append(image)
        self._getImages(disks)

    def _getImages(self, disks):

        for disk in disks:
            if disk[1] == CDROM_IMAGE:
                continue
            else:
                sd, image = disk
                imageDom = '%s:%s' % (sd, image)
                if imageDom not in self.imagesList.keys():
                    self.imagesList[imageDom] = []

                self._getVols(self.pool, sd, image)

    def _getVols(self, pool, sd, image):
        imageDom = '%s:%s' % (sd, image)
        volumes = self.serverConnection.getVolumesList(sd, pool, image)

        volParents = {}

        # An image may have multiple leaf volumes, eg previewing a snapshot
        leafvols = []

        for vol in volumes['uuidlist']:
            res = self.serverConnection.getVolumeInfo(sd, pool, image, vol)
            volParents[vol] = res['info']['parent']
            imageDomVol = '%s:%s:%s' % (sd, image, vol)
            self.volumesList[imageDomVol] = []

            if res['info']['voltype'] == 'LEAF':
                leafvols.append(vol)

            if res['info']['voltype'] == 'SHARED':
                leafvols.append(vol)
                self.volumesList[imageDomVol].append('Template')

        self.imagesList[imageDom] = self._buildVolumesChain(leafvols,
                                                            volParents, sd,
                                                            image)

    def _buildVolumesChain(self, leafvols, volParents, sd, image):

        volChain = []
        for leafvol in leafvols:
            if volChain:
                # insert an empty 'volume' to differentiate multiple chains
                volChain.insert(0, '')

            volChain.insert(0, leafvol)
            next = leafvol
            while next in volParents.keys():
                prev = next
                next = volParents[prev]
                if next in volParents.keys():
                    volChain.insert(0, next)
                else:
                    if next == BLANK_UUID:
                        app = 'Template independent'
                    else:
                        app = '%s' % (next)

                    imageDomVol = '%s:%s:%s' % (sd, image, prev)
                    templList = self.volumesList[imageDomVol]
                    if len(templList) == 0:
                        templList.append(app)

        return volChain

    def _buildStorageTable(self):
        self.domainsList = {}
        self.vmList = {}
        self.imagesList = {}
        self.volumesList = {}
        self.tbl = {}
        self.canAccessVms = False
        for header in self.columns:
            self.tbl[header] = []

        pools = self.serverConnection.getConnectedStoragePoolsList()
        if not pools['poollist']:
            return

        self.pool = pools['poollist'][0]

        # Check if its possible to access the ovf files
        try:
            files = os.listdir(os.path.join(DC, self.pool, MASTERSD_PATH))
            self.canAccessVms = True
        except (OSError, IOError):
            self.canAccessVms = False

        if self.canAccessVms:
            self._getVms(self.pool, files)
        else:
            self._getSds(self.pool)

        self._prepareTable()

    def _prepareTable(self):
        '''Prepare the table for printing'''

        if self.canAccessVms:
            colStart = 'Vms'
            startObj = self.vmList
        else:
            colStart = 'Domains'
            startObj = self.domainsList

        for obj, obj_data in startObj.items():
            row = 0

            obj_data = dict(obj_data)  # copy
            refs = obj_data.pop('Refs')
            self.tbl[colStart].append([obj] + obj_data.values())

            if len(refs) == 0:
                self.tbl['Images'].append([])
                self.tbl['Volumes'].append([])
                self.tbl['Template'].append([])

            for ref in refs:
                if self.canAccessVms:
                    sdUUID, image = ref
                    if row == 0:
                        sd_data = dict(self.domainsList[sdUUID])
                        del sd_data['Refs']
                        self.tbl['Domains'].append([sdUUID] + sd_data.values())
                    else:
                        self.tbl['Vms'].append([])
                        self.tbl['Domains'].append([])
                else:
                    sdUUID = obj
                    image = ref

                    if row == 0:
                        pass
                    else:
                        self.tbl['Domains'].append([])

                imageDom = '%s:%s' % (sdUUID, image)
                self.tbl['Images'].append([image])
                volsToAdd = self.imagesList[imageDom]
                self.tbl['Volumes'].append(volsToAdd)
                templToAdd = []
                for vol in volsToAdd:
                    domImageVol = '%s:%s:%s' % (sdUUID, image, vol)
                    templ = self.volumesList.get(domImageVol, [])
                    if templ:
                        templToAdd.append(templ[0])
                    else:
                        templToAdd.append('')
                self.tbl['Template'].append(templToAdd)
                row += 1

    def _printRowSep(self, rowSep, columnNum):
        sepLine = '+' + (rowSep + '+') * columnNum
        print sepLine

    def _printTableSegment(self, depth, alignCenter=False, printHeader=False):
        '''Printing the Table'''
        UUID_LEN = 36

        row = 0
        rows = 1
        rowSep = (UUID_LEN + 2) * '-'
        colStart = 0

        if not self.canAccessVms:
            colStart = 1
        columnNum = len(self.columns) - colStart

        if printHeader:
            self._printRowSep(rowSep, columnNum)

        while row <= rows:
            line = '|'

            for column in self.columns[colStart:]:
                if printHeader:
                    printObject = [column]
                else:
                    printObject = self.tbl[column][depth]

                if len(printObject) > rows:
                    rows = len(printObject)

                try:
                    adding = printObject[row]
                except IndexError:
                    adding = ''

                if alignCenter:
                    adding = adding.center(UUID_LEN)
                else:
                    adding = adding.ljust(UUID_LEN)

                line = line + ' %s |' % (adding)

            print '%s' % (line)
            row += 1

        self._printRowSep(rowSep, columnNum)

    def show(self):
        '''Show the Storage Table'''

        try:
            depth = 0
            self._printTableSegment(0, alignCenter=True, printHeader=True)

            for depth in range(0, len(self.tbl['Domains'])):
                self._printTableSegment(depth)

            return 0, ''
        except:
            return 1, traceback.format_exc()

if __name__ == '__main__':
    from vdsm import vdscli
    import sys

    try:
        rc, msg = StorageTable(vdscli.connect()).show()
    except socket.error as se:
        if se.errno != errno.ECONNREFUSED:
            raise se
        else:
            msg = 'Unable to connect to vdsmd ' + str(se)
            rc = 1
    if rc:
        print >>sys.stderr, msg
    sys.exit(rc)
